# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"array_split_divpoints_ntotal": "01_util.ipynb",
         "fftshift_checkerboard": "01_util.ipynb",
         "cartesian_aberrations": "01d_util_illumination.ipynb",
         "memory_mb": "01_util.ipynb",
         "memory_gb": "01_util.ipynb",
         "fourier_coordinates_2D": "01_util.ipynb",
         "array_split_divpoints": "01_util.ipynb",
         "R_factor": "01_util.ipynb",
         "distance": "01_util.ipynb",
         "rel_dist": "01_util.ipynb",
         "Param": "01_util.ipynb",
         "validate_standard_param": "01_util.ipynb",
         "format_standard_param": "01_util.ipynb",
         "asParam": "01_util.ipynb",
         "make_default": "01_util.ipynb",
         "PARAM_PREFIX": "01_util.ipynb",
         "single_sideband_reconstruction": "01_util.ipynb",
         "single_sideband_kernel": "01_util.ipynb",
         "sector_mask": "01_util.ipynb",
         "wavelength": "01_util.ipynb",
         "DOF": "01_util.ipynb",
         "dense_to_sparse_kernel": "01_util.ipynb",
         "advanced_raster_scan": "01_util.ipynb",
         "get_qx_qy_1D": "01_util.ipynb",
         "get_qx_qy_2D": "01_util.ipynb",
         "scatter_add_patches": "01_util.ipynb",
         "gather_patches": "01_util.ipynb",
         "HSV_to_RGB": "01a_util_plot.ipynb",
         "P1A_to_HSV": "01a_util_plot.ipynb",
         "imsave": "01a_util_plot.ipynb",
         "plot_complex_multi": "01a_util_plot.ipynb",
         "plot": "01a_util_plot.ipynb",
         "zplot": "01a_util_plot.ipynb",
         "plotAbsAngle": "01a_util_plot.ipynb",
         "center_of_mass_kernel": "01b_util_kernels.ipynb",
         "sparse_to_dense_datacube_kernel_crop": "01b_util_kernels.ipynb",
         "sparse_to_dense_datacube_crop": "01b_util_kernels.ipynb",
         "sparse_to_dense_datacube_crop_gain_mask_kernel": "01b_util_kernels.ipynb",
         "sparse_to_dense_datacube_crop_gain_mask": "01b_util_kernels.ipynb",
         "fftshift_kernel": "01b_util_kernels.ipynb",
         "fftshift_pad_kernel": "01b_util_kernels.ipynb",
         "virtual_annular_image_kernel": "01b_util_kernels.ipynb",
         "crop_symmetric_around_center_kernel": "01b_util_kernels.ipynb",
         "crop_symmetric_around_center": "01b_util_kernels.ipynb",
         "rotate_kernel": "01b_util_kernels.ipynb",
         "rotate": "01b_util_kernels.ipynb",
         "sum_kernel": "01b_util_kernels.ipynb",
         "sum_frames": "01b_util_kernels.ipynb",
         "rebin_kernel": "01b_util_kernels.ipynb",
         "disk_overlap_function": "01c_util_ssb.ipynb",
         "disk_overlap_kernel": "01c_util_ssb.ipynb",
         "double_overlap_intensitities_in_range": "01c_util_ssb.ipynb",
         "find_rotation_angle_with_double_disk_overlap": "01c_util_ssb.ipynb",
         "cartesian_aberrations_single": "01d_util_illumination.ipynb",
         "aperture": "01d_util_illumination.ipynb",
         "ZernikeProbe2": "01d_util_illumination.ipynb",
         "ZernikeProbeSingle": "01d_util_illumination.ipynb",
         "ZernikeProbe": "01d_util_illumination.ipynb",
         "sdebug": "01e_util_io.ipynb",
         "h5write": "01e_util_io.ipynb",
         "h5append": "01e_util_io.ipynb",
         "h5read": "01e_util_io.ipynb",
         "h5info": "01e_util_io.ipynb",
         "h5options": "01e_util_io.ipynb",
         "Metadata4D": "06_sparse_data.ipynb",
         "get_CoM": "06_sparse_data.ipynb",
         "get_probe_size": "06_sparse_data.ipynb",
         "Sparse4DData": "06_sparse_data.ipynb",
         "fasta2": "10a_fasta.ipynb",
         "setup_logging": "20_setup.ipynb",
         "configure_node": "20_setup.ipynb",
         "load_smatrix_data_list2": "20_setup.ipynb",
         "initial_probe_amplitude": "20_setup.ipynb",
         "prepare_beam_parameters": "20_setup.ipynb",
         "initial_smatrix": "20_setup.ipynb",
         "report_smatrix_parameters": "20_setup.ipynb",
         "report_initial_probes": "20_setup.ipynb",
         "calc_psi": "30_operators.ipynb",
         "AtF2": "30_operators.ipynb",
         "dS": "30_operators.ipynb",
         "A_realspace": "30_operators.ipynb",
         "calc_psi_denom": "30_operators.ipynb",
         "Qoverlap_real2": "30_operators.ipynb",
         "split": "30_operators.ipynb",
         "A": "30_operators.ipynb",
         "complex_matmul": "30_operators.ipynb",
         "A_fast_full2": "30_operators.ipynb",
         "A_fast_full3": "30_operators.ipynb",
         "A_fast_full5": "30_operators.ipynb",
         "A_fast_full4": "30_operators.ipynb",
         "AH_S": "30_operators.ipynb",
         "smatrix_phase_factorsBDK": "30_operators.ipynb",
         "smatrix_phase_factorsKB": "30_operators.ipynb",
         "smatrix_forward_kernel": "40_operators.kernels.ipynb",
         "smatrix_backward_kernel_S": "40_operators.kernels.ipynb",
         "phase_factor_kernelDBK": "40_operators.kernels.ipynb",
         "phase_factor_kernelKB": "40_operators.kernels.ipynb",
         "smatrix_forward_kernel_fast_full4": "40_operators.kernels.ipynb",
         "split_kernel": "40_operators.kernels.ipynb",
         "split_kernel4": "40_operators.kernels.ipynb",
         "split_kernel5": "40_operators.kernels.ipynb",
         "split_kernel2": "40_operators.kernels.ipynb",
         "split_kernel3": "40_operators.kernels.ipynb",
         "overlap_kernel_real2": "40_operators.kernels.ipynb",
         "psi_denom_kernel": "40_operators.kernels.ipynb",
         "psi_kernel": "40_operators.kernels.ipynb",
         "A_realspace_kernel": "40_operators.kernels.ipynb",
         "AtF2_kernel": "40_operators.kernels.ipynb",
         "dS_kernel": "40_operators.kernels.ipynb",
         "prox_D_gaussian_kernel": "60_loss_functions.ipynb",
         "prox_D_gaussian": "60_loss_functions.ipynb",
         "gradz_poisson_sparse_kernel": "60_loss_functions.ipynb",
         "gradz_poisson_sparse": "60_loss_functions.ipynb",
         "gradz_gaussian_sparse_kernel": "60_loss_functions.ipynb",
         "gradz_gaussian_sparse": "60_loss_functions.ipynb",
         "sparse_amplitude_loss_kernel": "60_loss_functions.ipynb",
         "sparse_amplitude_loss": "60_loss_functions.ipynb",
         "sparse_smooth_truncated_amplitude_loss_kernel": "60_loss_functions.ipynb",
         "sparse_smooth_truncated_amplitude_loss": "60_loss_functions.ipynb",
         "sparse_smooth_truncated_amplitude_prox_kernel": "60_loss_functions.ipynb",
         "sparse_smooth_truncated_amplitude_prox": "60_loss_functions.ipynb",
         "sparse_amplitude_prox_kernel": "60_loss_functions.ipynb",
         "sparse_amplitude_prox": "60_loss_functions.ipynb",
         "ReconstructionOptions": "90_core.ipynb",
         "SMeta": "90_core.ipynb"}

modules = ["torch_imports.py",
           "util.py",
           "core.py",
           "algorithm.py",
           "setup.py",
           "operators.py",
           "kernels.py"]

doc_url = "https://philipppelz.github.io//smpr3d/"

git_url = "https://github.com/s-matrix/smpr3d/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10a_fasta.ipynb (unless otherwise specified).

__all__ = ['fasta2']

# Cell

import torch as th
from timeit import default_timer as timer
import logging
from torch.utils.tensorboard import SummaryWriter
from torch.nn import Module
from .util import *
from .core import SMeta
import numpy as np
from .operators import smatrix_phase_factorsBDK

def fasta2(s_meta: SMeta, A, AH_S, AH_Psi, AH_r, prox_data, Psi_gen: Module,
           a: th.Tensor, S0: th.Tensor, Psi0, C0: th.Tensor, Ap0: th.Tensor, r0: th.Tensor, opts,
           S_sol: th.Tensor = None, Psi_sol: th.Tensor = None, r_sol: th.Tensor = None,
           summary: SummaryWriter = None):
    """
    Block coordinate descent for scanning diffraction experiments with the latent variables
    S: represents the sample, Psi: represents the incoming wave function, r: represents the scanning positions

    :param A:        multi-linear forward operator. must be callable with 3 parameters: A(S, Psi, r)
    :param AH_S:     adjoint of the linear forward operator A_S with all variables fixed but S. Must be callable with 3 parameters: AH_S(z, Psi, r)
    :param AH_Psi:   adjoint of the linear forward operator A_Psi with all variables fixed but Psi. Must be callable with 3 parameters: AH_Psi(z, S, r)
    :param AH_r:     adjoint of the linear forward operator A_r with all variables fixed but r. Must be callable with 3 parameters: AH_r(z, S, Psi)
    :param prox_data:proximal operator for the data term
    :param Psi_gen:      a function that creates the Fourier space probe from aberration coefficients C
    :param a:        float (D, K, MY, MX) measured amplitudes
    :param S0:       float (B_tile, NY, NX, 2) initial S-matrix
    :param C0:       float (12, D) initial aberration coefficients
    :param A0:       float (D, MY, MX) initial apertures
    :param r0:       float (D, K, 2) initial positions
    :param opts:     dictionary of optimization options
    :param S_sol:    optional, float (B_tile, NY, NX, 2) initial S-matrix
    :param Psi_sol:  optional, float (D, MY, MX, 2) initial probes
    :param r_sol:    optional,float (D, K, 2) initial positions
    :param summary:  optional, tensorboard summary writer, used to write intermediate results to tensorboard for vizualisation
    :return: tuple (S, Psi, r), outs, opts
                outs is a dictionary with the keys "R_factors"
                    if a S_sol is provided it has a key "S_errors"
                    if a Psi_sol is provided it has a key "Psi_errors"
    """
    tau_S = opts.tau_S
    tau_Psi = opts.tau_Psi
    tau_r = opts.tau_r

    # lambda operator update speed
    beta = opts.beta
    device = S0.device

    # Maximum number of updates
    max_iters = opts.max_iters
    rank = opts.node_config.rank
    world_size = opts.node_config.world_size
    verbose = opts.verbose

    # Get diffraction pattern shape
    _, _, MY, MX = a.shape

    S_errors = []
    R_factors = []
    Psi_errors = []
    r_errors = []

    # Move probe, probe positions and scattering matrix to the chosen device
    Ap = Ap0.to(device)

    C = C0.to(device)
    C.requires_grad = True

    r = r0.to(device)
    S = S0.to(device)
    if Psi0 is not None:
        Psi = Psi0.to(device)
    else:
        Psi = Psi_gen(C, Ap)

    # use same phase factors for forward and backward
    Psi.phase_factors = smatrix_phase_factorsBDK(Psi, r, s_meta.take_beams, s_meta.q_dft, s_meta.S_shape[0], out=None)

    if rank == 0 and verbose >= 2:
        logging.info(f"a[0,0]**2 sum  : {(a[0, 0] ** 2).sum()}")
        # logging.info(f"z_hat[0,0] norm: {norm(z_hat[0, 0]) ** 2}")

    if world_size == 1:
        ii = 1
        # rexitw = th.ifft(z_hat[:, ii], 2, True)
        # plot_complex_multi(complex_numpy(rexitw.cpu()), f'real space exit waves, position {ii}')

    z1 = th.fft.fft2(A(S, Psi, r), norm='ortho')
    a_model = th.abs(z1)

    # cb = th.as_tensor(fftshift_checkerboard(MY//2,MX//2), device=z1.device)
    # ww = th.fft.fftshift(z1[0,:9],(1,2)) * cb
    #
    # plot_complex_multi(ww.cpu().numpy(), f'real space exit waves, position {ii}')

    # print(a.shape, a_model.shape)
    fac = (1 - (a / (a_model + 1e-3)))
    z1.mul_(fac)
    dz = th.fft.ifft2(z1, norm='ortho')

    S_mom = None
    C_mom = None
    m = [MY // 4, MX // 4]

    S_accel1 = S
    z_accel1 = z1
    alpha1 = 1
    momentum = 0.5
    dampening = 0

    for i in range(max_iters):
        start = timer()

        dS = AH_S(dz, Psi, r)

        if rank == 0 and i == 0:
            logging.info(f'max_abs_dS: {th.max(th.abs(dS))}   max_abs_S: {th.max(th.abs(S.detach()))}')
        if S_mom is None:
            S_mom = th.clone(dS).detach()
        else:
            S_mom.mul_(momentum).add_(dS, alpha=1 - dampening)

        dS.add_(S_mom, alpha=momentum)
        # plotAbsAngle(dS[0, m[0]:-m[0], m[1]:-m[1]].cpu().numpy(), f'dS[{i}]')

        S.add_(dS, alpha=-tau_S)
        del dS
        z1 = th.fft.fft2(A(S, Psi, r), norm='ortho')

        R_fac = R_factor(z1, a, world_size)

        a_model = th.abs(z1)
        fac = (1 - (a / (a_model + 1e-3)))
        z1.mul_(fac)
        dz = th.fft.ifft2(z1, norm='ortho')

        del a_model
        del fac
        del z1

        # solve for probe (step 3 of algorithm 1 from paper)
        if opts.optimize_psi(i):
            # Calculate probe update
            dPsi = AH_Psi(dz, S, r)
            if rank == 0 and i == 0:
                logging.info(f'max_abs_dPsi: {th.max(th.abs(dPsi))}   max_abs_Psi: {th.max(th.abs(Psi.detach()))}')
            # Psi -= tau_Psi * dPsi
            Psi = Psi_gen(C, Ap)
            Psi.backward(dPsi)
            dC = C.grad
            if C_mom is None:
                C_mom = th.clone(C.grad).detach()
            else:
                C_mom.mul_(momentum).add_(C.grad, alpha=1 - dampening)
            # Apply probe update
            dC = dC.add(C_mom, alpha=momentum)
            C = C.detach().clone().add(dC, alpha=- tau_Psi)
            C.requires_grad = True
            if rank == 0:
                logging.info(f'C[0] = {C[0].detach().cpu().numpy()}, C.grad = {dC[0].cpu().numpy()}')
            Psi = Psi_gen(C, Ap)
            Psi.phase_factors = smatrix_phase_factorsBDK(Psi, r, s_meta.take_beams, s_meta.qf, s_meta.q_coords, s_meta.S_shape[0], out=None)

            if verbose == 2 and rank == 0 and summary is not None:
                summary.add_scalar(f'details/max_abs_dPsi', th.max(th.abs(dPsi)), i, timer())
                summary.add_scalar(f'details/max_abs_Psi', th.max(th.abs(Psi.detach())), i, timer())
            del dPsi

        # Logging
        if verbose > 0:
            end = timer()
            R_factors.append(R_fac.cpu().item())
            s = f"{i:03d}/{max_iters:03d} [{(end - start):-02.4f}s] R-factor: {R_factors[-1]:3.3g}   "
            if summary is not None:
                if verbose >= 1 and rank == 0:
                    summary.add_scalar(f'errors/R_factor', R_fac, i, timer())
            if Psi_sol is not None:
                err_Psi = rel_dist(Psi.detach(), Psi_sol)
                Psi_errors.append(err_Psi.cpu().item())
                s += f"err_Psi: {Psi_errors[-1]:3.3g}   "
                if summary is not None and rank == 0:
                    if verbose >= 1:
                        summary.add_scalar(f'errors/probe_error', err_Psi, i, timer())
            if S_sol is not None and rank == 0:
                S[th.isnan(S)] = 0

                err_S = rel_dist(S[:, m[0]:-m[0], m[1]:-m[1]], S_sol[:, m[0]:-m[0], m[1]:-m[1]])
                # plotcxmosaic(complex_numpy(
                #     complex_mul_conj(S[:16, m[0]:-m[0], m[1]:-m[1]], S0[:16, m[0]:-m[0], m[1]:-m[1]]).cpu()),
                #              f'S_sol[{0}]')
                # plotcxmosaic(complex_numpy(S[:16, m[0]:-m[0], m[1]:-m[1]].cpu()), f'S[{i}]')
                # plotAbsAngle(complex_numpy(S[0, m[0]:-m[0], m[1]:-m[1]].cpu() - S_sol[0, m[0]:-m[0], m[1]:-m[1]].cpu()), 'S - S_sol')
                S_errors.append(err_S.cpu().item())
                s += f"err_S: {S_errors[-1]:3.3g}"
                if summary is not None:
                    if verbose >= 1:
                        summary.add_scalar(f'errors/S_error', err_S, i, timer())
            if rank == 0:
                logging.info(s)

    if rank == 0:
        logging.info(f'{i} iterations finished.')

    outs = Param()
    outs.R_factors = np.asarray(R_factors)
    if S_sol is not None:
        outs.S_errors = np.asarray(S_errors)
    if Psi_sol is not None:
        outs.Psi_errors = np.asarray(Psi_errors)
    if r_sol is not None:
        outs.r_errors = np.asarray(r_errors)
    S = S.cpu()
    Psi = Psi.detach().cpu()
    r = r.cpu().numpy()
    C = C.detach().cpu().numpy()

    return (S, Psi, C, r), outs, opts


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/20_setup.ipynb (unless otherwise specified).

__all__ = ['setup_logging', 'configure_node', 'load_smatrix_data_list2', 'initial_probe_amplitude',
           'prepare_beam_parameters', 'initial_smatrix', 'report_smatrix_parameters', 'report_initial_probes']

# Cell
import logging
import sys

def setup_logging(path, log_filename):
    logFormatter = logging.Formatter("%(asctime)s %(message)s")
    rootLogger = logging.getLogger()
    fileHandler = logging.FileHandler("{0}/{1}.log".format(path, log_filename))
    fileHandler.setFormatter(logFormatter)
    rootLogger.addHandler(fileHandler)
    consoleHandler = logging.StreamHandler(sys.stdout)
    consoleHandler.setFormatter(logFormatter)
    rootLogger.addHandler(consoleHandler)
    rootLogger.setLevel(logging.INFO)

# Cell

import os
from .util import *
import torch as th
import torch.distributed as dist
from numba import cuda
import GPUtil
import psutil
import numpy as np
import h5py as h5

def configure_node(dist_backend, init_method):
    args = Param()
    args.dist_backend = dist_backend

    is_using_slurm = os.environ.get('SLURM_NTASKS') is not None
    if is_using_slurm:
        SLURM_LOCALID = int(os.environ.get('SLURM_LOCALID'))
        SLURM_PROCID = int(os.environ.get('SLURM_PROCID'))
        SLURM_NTASKS = int(os.environ.get('SLURM_NTASKS'))
        SLURM_NTASKS_PER_NODE = int(os.environ.get('SLURM_NTASKS_PER_NODE'))
        # logging.info(f'SLURM_LOCALID: {SLURM_LOCALID}')
        # logging.info(f'SLURM_PROCID: {SLURM_PROCID}')
        # logging.info(f'SLURM_NTASKS: {SLURM_NTASKS}')
        # logging.info(f'SLURM_NTASKS_PER_NODE: {SLURM_NTASKS_PER_NODE}')
        args.slurm_tasks_per_node = SLURM_NTASKS_PER_NODE if SLURM_NTASKS_PER_NODE is not None else 0
        args.rank = SLURM_PROCID if SLURM_PROCID is not None else 0
        args.gpu = SLURM_LOCALID if SLURM_LOCALID is not None else args.rank
        args.world_size = SLURM_NTASKS if SLURM_NTASKS is not None else 1
        args.is_distributed = True
        args.scheduler = 'slurm'
    else:
        args.rank = 0
        args.world_size = 1
        args.gpu = 0
        args.is_distributed = False
        args.scheduler = 'local'

    args.device = th.device(f'cuda:{args.gpu}')
    # if args.world_size > 1:
        # let numba know which device we are running the kernels on
    logging.info(f'rank {args.rank} avail gpus: {[x.id for x in GPUtil.getGPUs()]}')
    logging.info(f'Selecting device: {args.gpu}')
    cuda.select_device(args.gpu)
    dist.init_process_group(backend=args.dist_backend, rank=args.rank, world_size=args.world_size,
                            init_method=init_method)

    ram_gpu_free_GB = []
    ram_cpu_free_GB = psutil.virtual_memory().available / 2 ** 30
    gpus = GPUtil.getGPUs()
    gpu = gpus[args.gpu]
    ram_gpu_free_GB = gpu.memoryFree / 1000
    if args.rank == 0:
        logging.info(f'Scheduler: {args.scheduler}')
        logging.info(f'System resources:')
    logging.info(f'Rank {args.rank}    Free CPU RAM: {ram_cpu_free_GB} GB')
    # if args.world_size > 1:
    #     dist.barrier()
    logging.info(f'Rank {args.rank}    Free GPU RAM: {ram_gpu_free_GB} GB')
    # if args.world_size > 1:
    #     dist.barrier()
    logging.info(f'Rank {args.rank} is using device {args.gpu}/{len(gpus)}: {gpu.name} driver: v{gpu.driver}')
    # if args.world_size > 1:
    #     dist.barrier()
    return args

# Cell
from numpy.fft import fftshift

def load_smatrix_data_list2(fn, device, rank, world_size, fftshift_data=False, subset=None, r_max=None):
    """
    Load S-matrix reconstruction data. Required hdf5 keys:

    Abbreviations:
    D: number of scans/ aperture functions
    K: number of scan positions
    MY/MX: detector shape
    NY/NX: S-matrix shape

    lambda:                         electron wavelength in Angstrom                 scalar float
    alpha_rad:                      convergence angle in radians                    scalar float
    C:                              aberration coefficients                         (12, D) float
    k_max:                          real-space Nyquist resolution (half-period)     (2,) float
    specimen_thickness_angstrom:    appoximate thickness in Angstrom                scalar float
    vacuum_probe:                   an image of the vacuum beam                     (MY, MX) float
    data:                           diffraction data, fft_shifted                   (D, K, MY, MX) float or integer
    r:                              real-space positions, in pixel coordinates      (D, K, 2) float
    probe_fourier:                  initial aperture functions                      (D, MY, MX) complex

    Optional hdf5 keys:

    relative_shifts:                relative shifts of the positions for different aberrations  (D, 2) float
    S_target                        target S-matrix from simulation                             (B, NY, NX) complex
    Psi_target                      target aperture functions from simulation                   (D, MY, MX) complex
    r_target                        target positions from simulation, in pixel coordinates      (D, K, 2) float

    :param fn: file name. Must be a valid hdf5 file
    :param device: target device for all data
    :param rank: rank of the loading process
    :param world_size: world_size of the current program
    :return: all needed data
    """

    if rank == 0:
        logging.info(f'Now      loading data file {fn}')
    with h5.File(fn, 'r') as f:
        D = len(subset)
        lam = f['lambda'][0]
        # step = f['skip'][()]
        alpha_rad = f['alpha_rad'][0]

        if subset is None:
            C = th.from_numpy(f['C'][:, :]).to(device)
            subset = np.arange(D)
        else:
            C = th.from_numpy(f['C'][:, subset]).to(device)

        dx = th.from_numpy(f['d_sm'][:])
        specimen_thickness_angstrom = f['specimen_thickness_angstrom'][0]
        vac = f['vacuum_probe'][:, :]

        if fftshift_data:
            vac = fftshift(vac)
        vacuum_probe = th.from_numpy(vac).to(device)

        data = []
        for d in subset:
            data.append(f[f'data_{d}'][:, :, :])

        if fftshift_data:
            for d in range(D):
                data[d] = fftshift(data[d], (1, 2))

        data = np.array(data)

        r0 = []
        for d in subset:
            r0.append(f[f'r_{d}'][:, :])

        r0 = np.array(r0)

        try:
            r_rel = f['relative_shifts'][...]
            if rank == 0:
                logging.info(f'Relative shifts of datasets:')
                for rr in r_rel:
                    logging.info(f'     {rr}')
            for d in range(D):
                r0[d] += r_rel[d, :]
        except:
            pass

        y_min = 1e6
        x_min = 1e6
        for d in range(D):
            y_min1 = np.min(r0[d][:, 0])
            x_min1 = np.min(r0[d][:, 1])
            y_min = np.min([y_min, y_min1])
            x_min = np.min([x_min, x_min1])

        for d in range(D):
            r0[d] -= [y_min, x_min]

        K = data.shape[1]
        MY = data.shape[2]
        MX = data.shape[3]
        detector_shape = np.array([MY, MX])

        divpts = array_split_divpoints(data, world_size, 1)
        d = data[:, divpts[rank]:divpts[rank + 1], :, :]
        I_target = th.from_numpy(d).to(device)
        r = th.from_numpy(r0[:, divpts[rank]:divpts[rank + 1], :]).to(device)

        y_max = -1e6
        x_max = -1e6
        for d in range(D):
            y_max1 = np.max(r0[d][:, 0])
            x_max1 = np.max(r0[d][:, 1])
            y_max = np.max([y_max, y_max1])
            x_max = np.max([x_max, x_max1])
        y_min = 1e6
        x_min = 1e6
        for d in range(D):
            y_min1 = np.min(r0[d][:, 0])
            x_min1 = np.min(r0[d][:, 1])
            y_min = np.min([y_min, y_min1])
            x_min = np.min([x_min, x_min1])

        if rank == 0:
            logging.info(f'Position array boundaries: [{y_min}:{y_max},{x_min}:{x_max}]')
            if y_min < 0 or x_min < 0:
                logging.warning(f'y_min = {y_min}, x_min = {x_min}, NEGATIVE INDICES ARE NOT ALLOWED!')
        if rank == 0:
            logging.info(f'memory allocated: {th.cuda.memory_allocated()/1024**2} MB')

        K_rank = I_target.shape[1]

        S_sol = None
        Psi_sol = None
        r_sol = None
        try:
            S_sol = th.as_tensor(f['S_target'][:, :, :], dtype=th.complex64).to(device)
        except:
            S_sol = None

        try:
            Psi_sol = []
            for d in range(D):
                Psi_sol.append(th.as_tensor(f[f'Psi_target_{d}'][:, :], dtype=th.complex64).to(device))
        except:
            Psi_sol = None

        Psi_sol = th.stack(Psi_sol,0)

        try:
            r_sol = []
            for d in range(D):
                r_sol.append(th.from_numpy(f[f'r_target_{d}'][:, :]).to(device))
        except:
            r_sol = None

        r_sol = th.stack(r_sol,0)
        # Psi = []
        # for d in range(D):
        #     Psi.append(f[f'Psi0_{d}'][:, :])
        #
        # if fftshift_data:
        #     for d in range(D):
        #         Psi[d] = fftshift(Psi[d], (1, 2))
        #         cb = fftshift_checkerboard(MY // 2, MX // 2)
        #         Psi[d] *= cb
        #
        # Psi0 = []
        # for d in subset:
        #     Psi0.append(cx_from_numpy(Psi[d]).to(device).squeeze())

        # S-matrix lateral dimensions.
        # Ensure that the scattering matrix real space sampling is identical to
        # that implied by the maximum reciprocal lattice vector of the
        # diffraction pattern.
        MY_max = np.max(np.array(MY))
        MX_max = np.max(np.array(MX))

        NY = int((np.ceil((y_max + MY_max) / MY_max) * MY_max).item())
        NX = int((np.ceil((x_max + MX_max) / MX_max) * MX_max).item())
        fy = NY // MY_max
        fx = NX // MX_max
        if rank == 0:
            logging.info(f'Finished loading data file {fn}')

        # Enforce same data for all arrays
        da = th.float32
        vacuum_probe = vacuum_probe.type(da)

        r = r.type(da)
        C = C.type(da)

        # for d in subset:
        #     Psi0[d] = Psi0[d].type(da)

        if r_sol is not None:
            r_sol = r_sol.type(da)
        return lam, alpha_rad, C, dx, specimen_thickness_angstrom, vacuum_probe, D, K, K_rank, MY, MX, NY, NX, \
               fy, fx, detector_shape, r, I_target, y_max, x_max, y_min, x_min, S_sol, Psi_sol, r_sol

# Cell
def initial_probe_amplitude(vacuum_probe, I_target, world_size, rank):
    D, K, MY, MX = I_target.shape
    # D x My x Mx
    I_mean = th.sum(I_target, 1)
    if world_size > 1:
        dist.all_reduce(I_mean, op=dist.ReduceOp.SUM)
    I_mean /= K

    if rank == 0:
        logging.info(f'I_mean             :{th.sum(I_mean, (1, 2)).cpu().numpy()}')

    # dim: D x K_rank
    # total intensity per diffraction pattern
    I_tot = th.sum(I_target, (2, 3))

    # max intensity over all diffraction patterns
    I_max, I_max_inds = th.max(I_tot, 1)
    if world_size > 1:
        dist.all_reduce(I_max, op=dist.ReduceOp.MAX)

    # max intensity over all diffraction patterns, for each defocus
    # dim: D
    if rank == 0:
        logging.info(f'I_max              :{I_max.cpu().numpy()}')

    # dim: D
    I_init = vacuum_probe.unsqueeze(0).repeat((D, 1, 1))
    I_norm = I_init.norm(1, dim=(1, 2))
    fac = I_max / I_norm
    I_init *= fac[:, None, None]
    if rank == 0:
        logging.info(f'I_init norm        :{I_init.norm(1, dim=(1, 2)).cpu().numpy()}')
    A_init = th.sqrt(I_init)

# Cell
def prepare_beam_parameters(take_beams, q2, specimen_thickness_angstrom, alpha_rad, lam, max_phase_error,
                            use_full_smatrix, device):
    # number of beams
    B = th.sum(take_beams).item()
    beam_numbers = th.ones_like(take_beams, dtype=th.long, device=device) * -1
    beam_numbers[take_beams] = th.arange(0, B, device=device)

    if use_full_smatrix:
        # override tiling of the aperture
        reduction_factor = 1.0
        B_tile = B
        tile_order = beam_numbers
    else:
        raise NotImplementedError('coming soon')

    if reduction_factor == 1:
        tile_map = beam_numbers[take_beams]
        tile_number = beam_numbers
    else:
        tile_map = -1
        tile_number = -1

    return B, B_tile, tile_order, tile_number, tile_map

# Cell
from numpy.fft import fftfreq

def initial_smatrix(shape, q_space_tiling, device, is_unitary, include_plane_waves, initial_depth=0, lam=0, q2=0,
                    dtype=th.complex64, is_pinned=False):
    """

    :param shape:               (4-tuple) shape of the S-matrix
    :param q_space_tiling:      (2D)
    :param device:              torch.device
    :param is_unitary:          bool, if the S-matrix should be unitary
    :param include_plane_waves: bool, if the S-matrix should be in the plane-wave basis
    :param initial_depth:       float, z-depth of the initial S-matrix in Angstrom, gives quadratic phase offset to beams
    :param lam:                 float, wavelength in Angstrom
    :param q2:                  2D, wavevector squared
    :param dtype:               th.dtype
    :param is_pinned:           bool, create pinned memory for CPU
    :return: initialized S-matrix
    """
    B, NY, NX = shape
    MY, MX = q_space_tiling.shape
    fy, fx = NY // MY, NX // MX
    S = th.zeros(shape, dtype=dtype, device=device)
    if is_pinned:
        S = S.pin_memory()
    if initial_depth > 0:
        take_beams = q_space_tiling >= 0
        tile_map = q_space_tiling[take_beams]
        depth_init = th.exp(1j* -np.pi * q2 * lam * initial_depth).to(device)
        for b in range(B):
            S[b] = th.mean(depth_init[take_beams][tile_map == b], axis=0)
    else:
        depth_init = th.zeros(q2.shape)
    if include_plane_waves:
        qx, qy = np.meshgrid(fftfreq(MX), fftfreq(MY))
        q = np.array([qy, qx])
        q_dft = th.from_numpy(q).to(device).type(S.dtype)
        coords = th.from_numpy(fftshift(np.array(np.mgrid[-MY // 2:MY // 2, -MX // 2:MX // 2]), (1, 2))).to(device)
        for b in range(B):
            cur_beam = q_space_tiling == b
            cur_beam = cur_beam[None, ...].expand_as(coords)
            c = coords[cur_beam]
            cur_planewave = th.exp(2j * np.pi * (q_dft[0] * c[0] + q_dft[1] * c[1]))
            S[b] = S[b] * cur_planewave.repeat((fy, fx))
    # S += th.randn_like(S) * 1e-10
    if is_unitary:
        sy = NY // 2
        sx = NX // 2
        CS = S[:, sy:sy + MY, sx:sx + MX]
        css = th.norm(CS,p=2, dim=(1, 2))
        S /= css[:, None, None]
    return S, depth_init

# Cell
from .util import memory_mb
def report_smatrix_parameters(rank, world_size, a, S, B, D, K, MY, MX, NY, NX, fy, fx, B_tile, K_rank,
                              specimen_thickness_angstrom, depth_init, y_max, x_max, y_min, x_min):
    if rank == 0:
        nonzero_measurements = float(th.sum(a.cpu() > 0))
        y_size = (y_max - y_min)
        x_size = (x_max - x_min)
        unknowns = y_size * x_size * B_tile
        logging.info(f"B (number of beams)      = {B}")
        logging.info(f'B_tile (number of tiles) = {B_tile}')
        logging.info(f"K (number of positions)  = {K}")
        logging.info(f"D (number of aberrations)= {D}")
        logging.info(f"M (detector size)        = {MY, MX}")
        logging.info(f"N (S-matrix size)        = {NY, NX}")
        logging.info(f"f (PRISM factor)         = {fy, fx}")
        logging.info(f"")
        logging.info(f'tile reduction_factor    = {B_tile / B}')
        logging.info(f'#unknowns                = {unknowns:3.3g}')
        logging.info(f'#nonzero_measurements    = {nonzero_measurements:3.3g}')
        logging.info(f'variable oversampling    = {nonzero_measurements / unknowns:-2.2f}')
        logging.info(f"")
        logging.info(f"initial thickness est.   = {specimen_thickness_angstrom / 10} nm")

        size_smatrix_gb = memory_mb((B_tile, NY, NX, 2), th.float32)
        size_exitwaves_gb = memory_mb((K_rank, D, MY, MX, 2), th.float32)
        logging.info(f"")
        logging.info(f"Memory needed for S-matrix    : {size_smatrix_gb:-2.1f} MB")
        logging.info(f"Memory needed for measurements: {size_exitwaves_gb / 2:-2.1f} MB")
        logging.info(f"Memory needed for exitwaves   : {size_exitwaves_gb:-2.1f} MB")
        logging.info(f"")
        if world_size > 1:
            dist.barrier()
        logging.info(f"rank {rank:-3d}  K_rank  = {K_rank:-4d}")
    else:
        if world_size > 1:
            dist.barrier()
        logging.info(f"rank {rank:-3d} K_rank  = {K_rank:-4d}")

# Cell
def report_initial_probes(rank, world_size, Psi_gen, C, A_init, specimen_thickness_angstrom, q, lam,
                          alpha_rad, summary=None):
    C_exit = C.clone()
    C_exit[0] += specimen_thickness_angstrom

    Psi_model = Psi_gen(C, A_init).detach()
    psi_model = th.ifft(Psi_model, 2, True)
    D, MY, MX, _ = Psi_model.shape
    cb = fftshift_checkerboard(MY // 2, MX // 2)

    if rank == 0:
        logging.info(f"Psi_model norm: {th.norm(Psi_model[0]) ** 2}")

    Psi_exit = Psi_gen(C_exit, A_init).detach()
    psi_exit = th.ifft(Psi_exit, 2, True)

    if world_size == 1:
        fig_Psi = plot_complex_multi(cb * fftshift(Psi_model.cpu(), (1, 2)),
                                     'Fourier space entrance surface probes')
        plot_complex_multi(psi_model.cpu(), 'Real Space entrance surface probes')
        plot_complex_multi(cb * fftshift(Psi_exit.cpu(), (1, 2)), f'Fourier space exit surface probes')
        plot_complex_multi(psi_exit.cpu(), f'Real Space exit surface probes')

        if summary is not None:
            summary.add_figure('init/Psi_entrance', fig_Psi, 0, close=True)
    else:
        if rank == 0:
            fig_Psi = plot_complex_multi(cb * fftshift(Psi_model.cpu(), (1, 2)),
                                         'Fourier Space entrance surface probes', show=False)
            fig_psi = plot_complex_multi(psi_model.cpu(), 'Real Space entrance surface probes', show=False)
            fig_Psi_ex = plot_complex_multi(cb * fftshift(Psi_exit.cpu(), (1, 2)),
                                            f'Fourier    exit surface probes', show=False)
            fig_psi_ex = plot_complex_multi(psi_exit.cpu(), f'Real Space exit surface probes',
                                            show=False)
            if summary is not None:
                summary.add_figure('init/Psi_entrance', fig_Psi, 0, close=True)
                summary.add_figure('init/psi_entrance', fig_psi, 0, close=True)
                summary.add_figure('init/Psi_exit', fig_Psi_ex, 0, close=True)
                summary.add_figure('init/psi_exit', fig_psi_ex, 0, close=True)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/90_core.ipynb (unless otherwise specified).


from __future__ import annotations


__all__ = ['Metadata4D', 'get_CoM', 'get_probe_size', 'Sparse4DData', 'ReconstructionOptions', 'SMeta']

# Cell
from ncempy.io.dm import fileDM
from .torch_imports import *
from .util import *
from skimage.filters import gaussian
import numpy as np
import h5py
#nbdev_comment from __future__ import annotations
import sigpy as sp
from fastcore.basics import *
from numba import cuda

# Cell
class Metadata4D:

    __repr__= basic_repr(['scan_step', 'pixel_step', 'k_max', 'alpha_rad', 'rotation_deg', 'E_ev', 'wavelength',
                          'aberrations'])

    def __init__(self, E_ev=None, ):
        self.scan_step = np.array([0,0])
        self.k_max = np.array([0,0])
        self.alpha_rad = -1
        self.rotation_deg = 0
        self.E_ev = E_ev
        self.wavelength = -1
        self.aberrations = np.zeros((12,))
        self.pixel_step = np.array([0,0])

    @staticmethod
    def from_dm4_file(filename):
        m = Metadata4D()

        with fileDM(filename) as f:
            m.E_ev = f.allTags['.ImageList.2.ImageTags.Microscope Info.Voltage']
            m.scan_step = np.array(f.scale[-2:]) * 10
        m.wavelength = wavelength(m.E_ev)
        return m

    def to_h5(self, file_path, key):
        with h5py.File(file_path, 'a') as f:
            g = f.create_group(key)
            g.create_dataset('scan_step', data=self.scan_step)
            g.create_dataset('k_max', data=self.k_max)
            g.create_dataset('alpha_rad', data=self.alpha_rad)
            g.create_dataset('rotation_deg', data=self.rotation_deg)
            g.create_dataset('E_ev', data=self.E_ev)
            g.create_dataset('wavelength', data=self.wavelength)
            g.create_dataset('aberrations', data=self.aberrations)
            g.create_dataset('pixel_step', data=self.pixel_step)

    @staticmethod
    def from_h5(file_path, key):
        res = Metadata4D()
        with h5py.File(file_path, 'r') as f:
            g = f[key]
            res.scan_step = g['scan_step'][...]
            res.pixel_step = g['pixel_step'][...]
            res.k_max = g['k_max'][...]
            res.alpha_rad = g['alpha_rad'][()]
            res.rotation_deg = g['rotation_deg'][()]
            res.E_ev = g['E_ev'][()]
            res.wavelength = g['wavelength'][()]
            res.aberrations = g['aberrations'][...]
        return res


# Cell
#from py4DSTEM
def get_CoM(ar):
    """
    Finds and returns the center of mass of array ar.
    """
    nx, ny = np.shape(ar)
    ry, rx = np.meshgrid(np.arange(ny), np.arange(nx))
    tot_intens = np.sum(ar)
    xCoM = np.sum(rx * ar) / tot_intens
    yCoM = np.sum(ry * ar) / tot_intens
    return xCoM, yCoM

#from py4DSTEM
def get_probe_size(DP, thresh_lower=0.01, thresh_upper=0.99, N=100):
    """
    Gets the center and radius of the probe in the diffraction plane.

    The algorithm is as follows:
    First, create a series of N binary masks, by thresholding the diffraction pattern DP with a
    linspace of N thresholds from thresh_lower to thresh_upper, measured relative to the maximum
    intensity in DP.
    Using the area of each binary mask, calculate the radius r of a circular probe.
    Because the central disk is typically very intense relative to the rest of the DP, r should
    change very little over a wide range of intermediate values of the threshold. The range in which
    r is trustworthy is found by taking the derivative of r(thresh) and finding identifying where it
    is small.  The radius is taken to be the mean of these r values.
    Using the threshold corresponding to this r, a mask is created and the CoM of the DP times this
    mask it taken.  This is taken to be the origin x0,y0.

    Accepts:
        DP              (2D array) the diffraction pattern in which to find the central disk.
                        A position averaged, or shift-corrected and averaged, DP work well.
        thresh_lower    (float, 0 to 1) the lower limit of threshold values
        thresh_upper    (float, 0 to 1) the upper limit of threshold values
        N               (int) the number of thresholds / masks to use

    Returns:
        r               (float) the central disk radius, in pixels
        x0              (float) the x position of the central disk center
        y0              (float) the y position of the central disk center
    """
    thresh_vals = np.linspace(thresh_lower,thresh_upper,N)
    r_vals = np.zeros(N)

    # Get r for each mask
    DPmax = np.max(DP)
    for i in range(len(thresh_vals)):
        thresh = thresh_vals[i]
        mask = DP > DPmax*thresh
        r_vals[i] = np.sqrt(np.sum(mask)/np.pi)

    # Get derivative and determine trustworthy r-values
    dr_dtheta = np.gradient(r_vals)
    mask = (dr_dtheta <= 0) * (dr_dtheta >= 2*np.median(dr_dtheta))
    r = np.mean(r_vals[mask])

    # Get origin
    thresh = np.mean(thresh_vals[mask])
    mask = DP > DPmax*thresh
    x0,y0 = get_CoM(DP*mask)

    return r,x0,y0

class Sparse4DData:
    __repr__=basic_repr(['scan_dimensions', 'frame_dimensions'])
    def __init__(self):
        self.indices = None
        self.counts = None
        self.scan_dimensions = None
        self.frame_dimensions = None

    @staticmethod
    def from_4Dcamera_file(filename, slice=None):
        with h5py.File(filename, 'r') as f0:
            frames = f0['electron_events/frames'][:]
            scan_dimensions = (f0['electron_events/scan_positions'].attrs['Ny'],
                               f0['electron_events/scan_positions'].attrs['Nx'])
            frame_dimensions = np.array((576, 576))

        def unragged_frames_size(frames):
            mm = 0
            for ev in frames:
                if ev.shape[0] > mm:
                    mm = ev.shape[0]
            return mm

        def make_unragged_frames(frames, scan_dimensions):
            unragged_frame_size = unragged_frames_size(frames.ravel())
            fr_full = np.zeros((frames.ravel().shape[0], unragged_frame_size), dtype=np.int32)
            fr_full[:] = np.iinfo(fr_full.dtype).max
            for ii, ev in enumerate(frames.ravel()):
                fr_full[ii, :ev.shape[0]] = ev
            fr_full_4d = fr_full.reshape((*scan_dimensions, fr_full.shape[1]))
            fr_full_4d = fr_full_4d[:, :-1, :]
            return fr_full_4d

        d = Sparse4DData()
        d.indices = np.ascontiguousarray(make_unragged_frames(frames.ravel(), scan_dimensions))
        if slice is not None:
            d.indices = d.indices[slice]
        d.scan_dimensions = np.array(d.indices.shape[:2])
        d.frame_dimensions = frame_dimensions
        d.counts = np.zeros(d.indices.shape, dtype=np.bool)
        d.counts[d.indices != np.iinfo(d.indices.dtype).max] = 1

        return d

    def crop_symmetric_center_(self, center, max_radius = None):
        if max_radius is None:
            y_min_radius = np.min([center[0], self.frame_dimensions[0] - center[0]])
            x_min_radius = np.min([center[1], self.frame_dimensions[1] - center[1]])
            max_radius = np.min([y_min_radius, x_min_radius])
        inds = sp.to_device(self.indices,0)
        frame_dimensions = sp.to_device(self.frame_dimensions, 0)
        xp = sp.backend.get_array_module(inds)
        new_frames, new_frame_dimensions = crop_symmetric_around_center(inds,frame_dimensions, center, max_radius)
        print(f'old frames shape: {self.indices.shape}')
        print(f'new frames shape: {new_frames.shape}')
        print(f'old frames frame_dimensions: {self.frame_dimensions}')
        print(f'new frames frame_dimensions: {new_frame_dimensions}')
        self.indices = new_frames
        self.counts = np.zeros(self.indices.shape, dtype=np.bool)
        self.counts[self.indices != np.iinfo(self.indices.dtype).max] = 1
        self.frame_dimensions = new_frame_dimensions

    def crop_symmetric_center(self, center, max_radius = None, verbose = True):
        if max_radius is None:
            y_min_radius = np.min([center[0], self.frame_dimensions[0] - center[0]])
            x_min_radius = np.min([center[1], self.frame_dimensions[1] - center[1]])
            max_radius = np.min([y_min_radius, x_min_radius])
        inds = sp.to_device(self.indices,0)
        xp = sp.backend.get_array_module(inds)
        frame_dims = xp.array(self.frame_dimensions)
        new_frames, new_frame_dimensions = crop_symmetric_around_center(inds, frame_dims, center, max_radius)
        if verbose:
            print(f'old frames shape: {self.indices.shape}')
            print(f'new frames shape: {new_frames.shape}')
            print(f'old frames frame_dimensions: {self.frame_dimensions}')
            print(f'new frames frame_dimensions: {new_frame_dimensions}')
        res = Sparse4DData()
        res.indices = new_frames
        res.counts = np.zeros(self.indices.shape, dtype=np.bool)
        res.counts[self.indices != np.iinfo(self.indices.dtype).max] = 1
        res.frame_dimensions = new_frame_dimensions
        res.scan_dimensions = self.scan_dimensions.copy()
        return res

    def rotate_(self, angle_rad, center=None):
        if center is None:
            center = self.frame_dimensions // 2
        new_indices = rotate(self.indices, self.frame_dimensions, center, angle_rad)
        self.indices = new_indices

    def rotate(self, angle_rad, center=None):
        if center is None:
            center = self.frame_dimensions // 2
        new_indices = rotate(self.indices, self.frame_dimensions, center, angle_rad)
        res = Sparse4DData()
        res.indices = new_indices
        res.counts = self.counts.copy()
        res.frame_dimensions = self.frame_dimensions
        res.scan_dimensions = self.scan_dimensions.copy()
        return res

    def sum_diffraction(self):
        res = sum_frames(self.indices, self.counts, self.frame_dimensions)
        return res

    @staticmethod
    def _determine_center_and_radius(data : Sparse4DData, manual=False, size=25, threshold=3e-1):
        sh = np.concatenate([data.scan_dimensions,data.frame_dimensions])
        c = np.zeros((2,))
        c[:] = (sh[-1] // 2, sh[-2] // 2)
        radius = np.ones((1,)) * sh[-1] // 2
        inds = sp.to_device(data.indices[:size, :size].astype(np.uint32),0)
        cts = sp.to_device(data.counts[:size, :size].astype(np.uint32),0)
        xp = sp.backend.get_array_module(inds)
        dc_subset = sparse_to_dense_datacube_crop(inds,cts, (size,size), data.frame_dimensions, c, radius, bin=2)
        dcs = xp.sum(dc_subset, (0, 1))
        m1 = dcs.get()
        m = (gaussian(m1.astype(np.float32),2) > m1.max() * threshold).astype(np.float)
        r, y0, x0 = get_probe_size(m)
        return 2 * np.array([y0,x0]), r*2

    def determine_center_and_radius(self, manual=False, size=25, threshold=3e-1):
        return Sparse4DData._determine_center_and_radius(self, manual, size=size, threshold=threshold)

    def to_dense(self, bin_factor, n_batches=4):
        dense = sparse_to_dense_datacube_crop_gain_mask(self.indices, self.counts.astype(np.int16), self.scan_dimensions,
                                                self.frame_dimensions, self.frame_dimensions/2, self.frame_dimensions[0]/2,
                                                self.frame_dimensions[0]/2, binning=bin_factor, n_batches=n_batches, fftshift=False)
        return dense

    @staticmethod
    def from_dense(dense, make_float = False, n_batches=1):
        res = Sparse4DData()
        res.frame_dimensions = np.array(dense.shape[-2:])
        res.scan_dimensions = np.array(dense.shape[:2])

        cuda.select_device(1)
        dev = th.device('cuda:1')

        inds = np.prod(res.frame_dimensions)
        if inds > 2**31:
            dtype = th.int64
        elif inds > 2**15:
            dtype = th.int32
        elif inds > 2**8:
            dtype = th.int16
        else:
            dtype = th.uint8

        nonzeros = np.sum((dense > 0),(2,3))
        nonzeros = np.max(nonzeros)

        bits_counts = np.log2(dense.max())
        if make_float:
            dtype_counts = th.float32
        else:
            if bits_counts > np.log2(2**31-1):
                dtype_counts = th.int64
            elif bits_counts > np.log2(2**15-1):
                dtype_counts = th.int32
            elif bits_counts > 8:
                dtype_counts = th.int16
            else:
                dtype_counts = th.uint8

        print(f'Using dtype: {dtype} for indices')
        print(f'Using dtype: {dtype_counts} for counts')

        threadsperblock = (16, 16)
        blockspergrid = tuple(np.ceil(res.scan_dimensions / threadsperblock).astype(np.int))

        ds = dense.shape
        K = dense.shape[0]
        divpts = array_split_divpoints_ntotal(K, n_batches)

        dense = th.as_tensor(dense)
        indices = th.zeros((*dense.shape[:2], nonzeros), dtype=dtype)
        counts = th.zeros((*dense.shape[:2], nonzeros), dtype=dtype_counts)

        indices[:] = th.iinfo(dtype).max
        fd = th.tensor(res.frame_dimensions).to(dev)
        for b in range(n_batches):

            denseb = dense[divpts[b]:divpts[b + 1]].to(dev)
            indicesb = indices[divpts[b]:divpts[b + 1]].to(dev)
            indicesb[:] = th.iinfo(dtype).max
            countsb = counts[divpts[b]:divpts[b + 1]].to(dev)

            dense_to_sparse_kernel[blockspergrid, threadsperblock](denseb, indicesb, countsb, fd)

            counts[divpts[b]:divpts[b + 1]] = countsb.cpu()
            indices[divpts[b]:divpts[b + 1]] = indicesb.cpu()

        res.indices = indices.numpy()
        res.counts = counts.numpy()

        print(f'frame_dimensions: {res.frame_dimensions}')
        print(f'scan_dimensions : {res.scan_dimensions}')

        cuda.select_device(0)

        return res

    @staticmethod
    def rebin(sparse_data: Sparse4DData, bin_factor : int, n_batches=4) -> Sparse4DData:
        dense = sparse_to_dense_datacube_crop_gain_mask(sparse_data.indices, sparse_data.counts.astype(np.int16), sparse_data.scan_dimensions,
                                                sparse_data.frame_dimensions, sparse_data.frame_dimensions/2, sparse_data.frame_dimensions[0]/2,
                                                sparse_data.frame_dimensions[0]/2, binning=bin_factor, n_batches=n_batches, fftshift=False)
        sparse = Sparse4DData.from_dense(dense)
        return sparse

    @staticmethod
    def fftshift(sparse_data: Sparse4DData) -> Sparse4DData:
        indices = sparse_data.indices
        scan_dimensions = sparse_data.scan_dimensions
        frame_dimensions = sparse_data.frame_dimensions
        center_frame = frame_dimensions / 2

        threadsperblock = (16, 16)
        blockspergrid = tuple(np.ceil(np.array(indices.shape[:2]) / threadsperblock).astype(np.int))

        no_count_indicator = np.iinfo(indices.dtype).max

        inds = sp.to_device(indices, 0)
        center_frame = sp.to_device(center_frame, 0)
        scan_dimensions = sp.to_device(scan_dimensions, 0)

        fftshift_kernel[blockspergrid, threadsperblock](inds, center_frame, scan_dimensions, no_count_indicator)

        res = Sparse4DData()
        res.indices = inds.get()
        res.counts = sparse_data.counts.copy()
        res.scan_dimensions = sparse_data.scan_dimensions.copy()
        res.frame_dimensions = sparse_data.frame_dimensions.copy()

        return res

    @staticmethod
    def fftshift_and_pad_to(sparse_data: Sparse4DData, pad_to_frame_dimensions) -> Sparse4DData:
        indices = sparse_data.indices
        scan_dimensions = sparse_data.scan_dimensions
        frame_dimensions = sparse_data.frame_dimensions
        center_frame = frame_dimensions / 2

        threadsperblock = (16, 16)
        blockspergrid = tuple(np.ceil(np.array(indices.shape[:2]) / threadsperblock).astype(np.int))

        no_count_indicator_old = np.iinfo(indices.dtype).max
        center_frame = sp.to_device(center_frame,0)
        xp = sp.backend.get_array_module(center_frame)

        inds = np.prod(pad_to_frame_dimensions)
        if inds > 2**15:
            dtype = xp.int64
        elif inds > 2**15:
            dtype = xp.int32
        elif inds > 2**8:
            dtype = xp.int16
        else:
            dtype = xp.uint8

        no_count_indicator_new = xp.iinfo(dtype).max

        inds = sp.to_device(indices,0).astype(dtype)
        pad_to_frame_dimensions = sp.to_device(pad_to_frame_dimensions,0)

        scan_dimensions = sp.to_device(scan_dimensions,0)

        fftshift_pad_kernel[blockspergrid, threadsperblock](inds, center_frame, scan_dimensions,  pad_to_frame_dimensions,
                                                            no_count_indicator_old, no_count_indicator_new)
        sparse_data.indices = inds.get()
        sparse_data.frame_dimensions = pad_to_frame_dimensions.get()
        return sparse_data

    def fftshift_(self):
        indices = self.indices
        scan_dimensions = self.scan_dimensions
        frame_dimensions = self.frame_dimensions
        center_frame = frame_dimensions / 2

        threadsperblock = (16, 16)
        blockspergrid = tuple(np.ceil(np.array(indices.shape[:2]) / threadsperblock).astype(np.int))

        no_count_indicator = np.iinfo(indices.dtype).max

        inds = sp.to_device(indices, 0)
        center_frame = sp.to_device(center_frame, 0)
        scan_dimensions = sp.to_device(scan_dimensions, 0)

        fftshift_kernel[blockspergrid, threadsperblock](inds, center_frame, scan_dimensions, no_count_indicator)
        self.indices = inds.get()
        return self

    def fftshift_and_pad_to_(self, pad_to_frame_dimensions):
        return Sparse4DData.fftshift_and_pad_to(self, pad_to_frame_dimensions)

    def bin(self, binning_factor, n_batches=4):
        res = Sparse4DData.rebin(self, binning_factor, n_batches=n_batches)
        return res

    def virtual_annular_image(self, inner_radius, outer_radius, center):
        inds = sp.to_device(self.indices, 0)
        xp = sp.backend.get_array_module(inds)
        cts = xp.array(self.counts, dtype=xp.uint32)
        ctr = xp.array(center)
        frame_dims = xp.array(self.frame_dimensions)
        img = xp.zeros(tuple(self.scan_dimensions), dtype=xp.uint32)
        no_count_indicator = xp.iinfo(self.indices.dtype).max
        threadsperblock = (16, 16)
        blockspergrid = tuple(np.ceil(np.array(self.indices.shape[:2]) / threadsperblock).astype(np.int))
        virtual_annular_image_kernel[blockspergrid, threadsperblock](img, inds, cts, inner_radius, outer_radius, ctr,
                                                                     frame_dims, no_count_indicator)
        return img.get()

    def fluence(self, dr):
        sum_electrons = self.counts.sum()
        area = np.prod(self.scan_dimensions) * dr**2
        return sum_electrons/area

    def flux(self, dr, dwell_time):
        fluence = self.fluence(dr)
        flux = fluence / (np.prod(self.scan_dimensions) * dwell_time)
        return flux

    def slice(self, slice):
        res = Sparse4DData()
        res.indices = np.ascontiguousarray(self.indices[slice])
        res.counts = np.ascontiguousarray(self.counts[slice])
        res.scan_dimensions = np.array(res.counts.shape[:2])
        res.frame_dimensions = self.frame_dimensions.copy()
        return res

    def center_of_mass(self: Sparse4DData):
        qx, qy = np.meshgrid(np.arange(self.scan_dimensions[0]),np.arange(self.scan_dimensions[1]))
        comx = th.zeros(self.scan_dimensions, dtype=th.float32)
        comy = th.zeros(self.scan_dimensions, dtype=th.float32)

        no_count_indicator = np.iinfo(self.indices.dtype).max

        mass = np.sum(self.counts,2)

        threadsperblock = (16, 16)
        blockspergrid = tuple(np.ceil(np.array(self.indices.shape[:2]) / threadsperblock).astype(np.int))

        qx = th.tensor(qx).astype(th.float32)
        qy = th.tensor(qy).astype(th.float32)
        center_of_mass_kernel[blockspergrid, threadsperblock](comx, comy, th.tensor(self.indices), th.tensor(self.counts.astype(np.uint32)),
                                                              th.tensor(self.frame_dimensions), no_count_indicator, qx, qy)
        comy = comy.get()
        comx = comx.get()
        comx /= mass + 1e-6
        comy /= mass + 1e-6
        comy[comy==0] = np.mean(comy[comy!=0])
        comx[comx==0] = np.mean(comx[comx!=0])
        comx -= np.mean(comx)
        comy -= np.mean(comy)
        return comy, comx

    def to_h5(self, file_path, key):
        with h5py.File(file_path, 'a') as f:
            g = f.create_group(key)
            g.create_dataset('counts', data=self.counts, compression="gzip", compression_opts=7)
            g.create_dataset('indices', data=self.indices, compression="gzip", compression_opts=7)
            g.create_dataset('frame_dimensions', data=self.frame_dimensions)
            g.create_dataset('scan_dimensions', data=self.scan_dimensions)

    @staticmethod
    def from_h5(file_path, key):
        res = Sparse4DData()
        with h5py.File(file_path, 'r') as f:
            g = f[key]
            res.counts = g['counts'][...]
            res.indices = g['indices'][...]
            res.frame_dimensions = g['frame_dimensions'][...]
            res.scan_dimensions = g['scan_dimensions'][...]
        return res

# Cell

class ReconstructionOptions:
    def __init__(self):
        pass

# Cell
from .util import fourier_coordinates_2D
import torch as th
import numpy as np
from numpy.fft import fftshift

class SMeta:
    def __init__(self, take_beams, dx, S_shape, MY, MX, device):
        self.f = S_shape[1:] / np.array([MY, MX])
        self.S_shape = S_shape
        self.q = th.as_tensor(fourier_coordinates_2D(S_shape[1:], dx.numpy(), centered=False), device=device)
        self.qf = th.as_tensor(fourier_coordinates_2D([MY, MX], dx.numpy(), centered=False), device=device)
        self.q2 = th.norm(self.q, dim=0) ** 2
        self.qf2 = th.norm(self.qf, dim=0) ** 2
        self.q_coords = th.from_numpy(
            fftshift(np.array(np.mgrid[-MY // 2:MY // 2, -MX // 2:MX // 2]), (1, 2))).to(device)
        self.r_indices = th.from_numpy(np.mgrid[:MY, :MX]).to(device)
        self.take_beams = take_beams.to(device)

        B = th.sum(take_beams).item()
        self.beam_numbers = th.ones_like(take_beams, dtype=th.long, device=device) * -1
        self.beam_numbers[take_beams] = th.arange(0, B, device=device)
        self.q_b = th.stack([self.qf[0,take_beams],self.qf[1,take_beams]]).T
        self.q_b_coords = th.from_numpy(np.mgrid[:MY, :MX]).to(device)
        self.q_dft = th.from_numpy(fourier_coordinates_2D([MY, MX], [1, 1], centered=False)).to(device)

        #S_shape          array (3,)
        #q              (NY, NX) x
        #qf             (MY, MX) x
        #q2             (NY, NX) x
        #q2f            (MY, MX) x
        #f              (2,)
        #q_coords       (MY, MX)
        #r_indices      (NY, NX)
        #take_beams     (MY, MX)
        #beam_numbers   (MY, MX)
        #q_b            (B, 2)
        #q_b_coords     (B, 2)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/50_functional.ipynb (unless otherwise specified).

__all__ = ['SMatrixSubpix', 'SparseSmoothTruncatedAmplitudeLoss', 'SparseSmoothTruncatedAmplitudeProx',
           'SparseAmplitudeLoss']

# Cell
from .torch_imports import *
from .util import cartesian_aberrations, fftshift_checkerboard
import numpy as np
import torch as th
from .operators import AtF2, A_realspace, dS
from numpy.fft import fftshift, fftfreq
from torch.autograd import Function
from .operators import split
from .util import sparse_amplitude_loss, plotAbsAngle, sparse_smooth_truncated_amplitude_loss, sparse_smooth_truncated_amplitude_prox

class SMatrixSubpix(th.autograd.Function):
    @staticmethod
    def forward(ctx, S: th.Tensor, psi: th.Tensor, r: th.Tensor) -> th.Tensor:
        """
        :param S: BB x N1 x N2       tensor
        :param psi: BB X M1 x M2     tensor in fourier space
        :param pos: K x 2 real                  tensor

        :return: D x K x M1 x M2                measured amplitudes in the far field
        """
        B, MY, MX = psi.shape
        K, _ = r.shape
        qy, qx = np.meshgrid(fftfreq(MY), fftfreq(MX), indexing='ij')
        # 2 x M1 x M2
        q = th.stack([th.as_tensor(qy), th.as_tensor(qx)]).float().cuda()
        rs = r - r.int()
        r_int = r.long()
        # K x MY x MX
        ramp = th.exp(-2j * np.pi *
            (q[0][None, None, ...] * rs[:, 0][:, None, None] + q[1][None, None, ...] * rs[:, 1][:, None, None]))

        # B x K x M1 x M2
        frames_exit = split(S.detach(), r_int, MY, MX)
        # plotAbsAngle(frames_exit[0, 0].detach().cpu(),'frames_exit')
        # B x K x M x M
        #subpixel shift the whole batch
        # print(f'psi norm: {th.norm(psi)**2}')
        Psi = th.fft.fft2(psi, norm='ortho')
        Psi = Psi.unsqueeze(1)
        Psi = Psi.repeat(1, ramp.shape[0], 1, 1)
        psi = th.fft.ifft2(Psi * ramp, norm='ortho')
        # print(f'psi norm: {th.norm(psi[0,0])**2}')
        # psi = th.ifft(Psi, 2, True)
        # plotAbsAngle(psi[0,0].detach().cpu(),'psi input')
        psi_exit = th.zeros((K, MY, MX), dtype=th.complex64, device=S.device)
        psi_exit = A_realspace(r_int, S.detach(), psi, psi_exit)
        # print(f'psi_exit norm: {th.norm(psi_exit[0])**2}')
        # plotAbsAngle(psi_exit[0].detach().cpu(), 'psi_exit')
        Psi_exit = th.fft.fft2(psi_exit, norm='ortho')
        # plotAbsAngle(Psi_exit[0].detach().cpu(), 'Psi_exit')
        # K x M x M
        a_model = th.abs(Psi_exit)
        # plot(a_model[0].cpu().detach().numpy(),'a_model')
        ctx.save_for_backward(frames_exit, Psi_exit, psi_exit, psi, r_int, ramp, th.Tensor([S.shape])[0].int())
        return a_model

    def backward(ctx, grad_output):
        frames_exit, Psi_exit, psi_exit, psi, r_int, ramp, ss = ctx.saved_tensors
        # (K x M x M)  * (K x M x M)
        grad_Psi = Psi_exit * grad_output
        # print(f'grad_Psi.shape {grad_output.shape}')
        # plotAbsAngle(fftshift(grad_Psi[0].cpu().numpy()), 'grad_Psi')
        # K x M x M
        grad_Psi = th.fft.ifft2(grad_Psi, norm='ortho')

        # print(f'grad_Psi.shape {grad_output.shape}')
        # plotAbsAngle(grad_Psi[0].cpu().numpy(),'grad_Psi')

        if ctx.needs_input_grad[0]:
            grad_S = th.zeros(tuple(ss.numpy()), device=Psi_exit.device, dtype=th.complex64)
            # print(Psi_exit.shape, grad_Psi.shape, psi.shape, r.shape, grad_S.shape)

            psi_int = th.abs(psi)**2
            psi_int_max = th.max(th.max(psi_int,-1).values,-1).values
            grad_S = dS(grad_Psi, psi_exit, psi, psi_int, psi_int_max, r_int, grad_S, alpha=0.1)
            # AtF2(grad_Psi, psi, r_int, grad_S)
        else:
            grad_S = None

        if ctx.needs_input_grad[1]:
            # (B x K x MY x MX) * (B x K x MY x MX)
            frames_exit_int_max = th.max(th.abs(frames_exit)**2)
            grad_psi = grad_Psi * frames_exit.conj() / frames_exit_int_max
            grad_psi = th.fft.fft2(grad_psi, norm='ortho')
            ramp = th.view_as_real(ramp)
            ramp[..., 1] *= -1
            ramp = th.view_as_complex(ramp)
            #(B x K x MY x MX) * (K x MY x MX)
            grad_psi = th.fft.ifft2(grad_psi * ramp, norm='ortho')
            #(B x MY x MX)
            grad_psi = th.mean(grad_psi, 1)
            # plotAbsAngle(grad_psi[0].cpu().numpy(), 'grad_psi')
            # grad_Psi = th.fft.ifft2(grad_psi, norm='ortho')
        else:
            grad_psi = None

        if ctx.needs_input_grad[2]:
            grad_pos = None
        else:
            grad_pos = None

        return grad_S, grad_psi, grad_pos

class SparseSmoothTruncatedAmplitudeLoss(Function):

    @staticmethod
    def forward(ctx, a_model, indices_target, counts_target):
        frame_dimensions = th.as_tensor(a_model.shape[1:], device=a_model.device)
        loss, grad = sparse_smooth_truncated_amplitude_loss(a_model, indices_target, counts_target, frame_dimensions)
        loss.requires_grad = True
        ctx.save_for_backward(grad)
        return loss

    @staticmethod
    def backward(ctx, *grad_outputs):
        grad_input, = ctx.saved_tensors
        grad_indices = None
        grad_counts = None
        return grad_input, grad_indices, grad_counts

class SparseSmoothTruncatedAmplitudeProx(Function):

    @staticmethod
    def forward(ctx, a_model, indices_target, counts_target):
        frame_dimensions = th.as_tensor(a_model.shape[1:], device=a_model.device)
        loss, grad = sparse_smooth_truncated_amplitude_prox(a_model, indices_target, counts_target, frame_dimensions)
        loss.requires_grad = True
        ctx.save_for_backward(grad)
        return loss

    @staticmethod
    def backward(ctx, *grad_outputs):
        grad_input, = ctx.saved_tensors
        grad_indices = None
        grad_counts = None
        return grad_input, grad_indices, grad_counts

class SparseAmplitudeLoss(Function):

    @staticmethod
    def forward(ctx, a_model, indices_target, counts_target):
        frame_dimensions = th.as_tensor(a_model.shape[1:], device=a_model.device)
        loss, grad = sparse_amplitude_loss(a_model, indices_target, counts_target, frame_dimensions)
        loss.requires_grad = True
        ctx.save_for_backward(grad)
        return loss

    @staticmethod
    def backward(ctx, *grad_outputs):
        grad_input, = ctx.saved_tensors
        grad_indices = None
        grad_counts = None
        return grad_input, grad_indices, grad_counts



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/00_torch_imports.ipynb (unless otherwise specified).

__all__ = []

# Cell
import torch as th
from torch import as_tensor,Tensor,ByteTensor,LongTensor,FloatTensor,HalfTensor,DoubleTensor
import torch.nn as nn
import torch.nn.functional as F
import torch.distributed as dist


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/40_operators.kernels.ipynb (unless otherwise specified).

__all__ = ['smatrix_forward_kernel', 'smatrix_backward_kernel_S', 'phase_factor_kernelDBK', 'phase_factor_kernelKB',
           'smatrix_forward_kernel_fast_full4', 'split_kernel', 'split_kernel4', 'split_kernel5', 'split_kernel2',
           'split_kernel3', 'overlap_kernel_real2', 'psi_denom_kernel', 'psi_kernel', 'A_realspace_kernel',
           'AtF2_kernel', 'dS_kernel']

# Cell
import numba.cuda as cuda
import cmath as cm

@cuda.jit
def smatrix_forward_kernel(S, phase_factors, rho, r_min, out):
    """
    :param S:               B x NY x NX
    :param phase_factors:   B x D x K x 2
    :param rho:               D x K x 2
    :param out:             D x K x MY x MX
    :return: exit waves in out
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    D, K, MY, MX, _ = out.shape
    B = S.shape[0]
    MM = MY * MX

    b = int(n // MM)
    my = (n - b * MM) // MX
    mx = (n - b * MM - my * MX)

    if n < B * MY * MX:
        for d in range(D):
            for k in range(K):
                # indexing with pixel precision
                rho0 = int(rho[d, k, 0] - r_min[0])
                rho1 = int(rho[d, k, 1] - r_min[1])

                a = S[b, rho0 + my, rho1 + mx, 0]
                c = S[b, rho0 + my, rho1 + mx, 1]
                u = phase_factors[b, d, k, 0]
                v = phase_factors[b, d, k, 1]

                val_real = a * u - c * v
                val_imag = c * u + a * v

                cuda.atomic.add(out, (d, k, my, mx, 0), val_real)
                cuda.atomic.add(out, (d, k, my, mx, 1), val_imag)

# Cell
@cuda.jit
def smatrix_backward_kernel_S(z, phase_factors, mean_probe_intensities, r, r_min, out, tau):
    """
    S-matrix has beam tilts included, pre-calculated scanning phase factors.
    Fastest to compute

    :param z:                       D x K x MY x MX x 2
    :param phase_factors:           B x D x K x 2
    :param r:                       D x K x 2
    :param mean_probe_intensities:  D
    :param out:                     B x NY x NX x 2
    :param z_strides:  (4,)
    :return: exit waves in out
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    D, K, MY, MX, _ = z.shape
    B = out.shape[0]

    b = n // (MY * MX)
    my = (n - b * (MX * MY)) // MX
    mx = (n - b * (MX * MY) - my * MX)

    if n < MY * MX * B:
        for d in range(D):
            for k in range(K):
                r0 = int(r[d, k, 0] - r_min[0])
                r1 = int(r[d, k, 1] - r_min[1])

                a = z[d, k, my, mx, 0]
                c = z[d, k, my, mx, 1]
                u = phase_factors[b, d, k, 0]
                v = phase_factors[b, d, k, 1]

                val_real = a * u + c * v
                val_imag = c * u - a * v

                val_real *= tau[0] / mean_probe_intensities[d]
                val_imag *= tau[0] / mean_probe_intensities[d]

                cuda.atomic.add(out, (b, r0 + my, r1 + mx, 0), val_real)
                cuda.atomic.add(out, (b, r0 + my, r1 + mx, 1), val_imag)

# Cell
@cuda.jit
def phase_factor_kernelDBK(Psi, rho, qB, out):
    """
    Calculate the phase factors (due to beam scan) probe wave function so that
    the probe is scanned to the correct place for each diffraction pattern

    :param Psi:         D x B
        Probe wave functions Fourier coefficient for each beam to be mutliplied
        by phase factor to account for beam scan position
    :param rho:           D x K x 2
        Probe positions in pixels
    :param qB:          2 x B
        Fourier space coordinates of the beams
    :param out:         D x B x K x 2
        Phase factors output
    :param out_strides: (3,)
    :return: scanning phases for all defoc, beams, positions
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    B, D, K, c = out.shape
    N = B * K * D

    b = n // (K * D)
    d = (n - b * (K * D)) // K
    k = (n - b * (K * D) - d * K)

    if n < N:
        rho0 = rho[d, k, 0]
        rho1 = rho[d, k, 1]
        Psic = Psi[d, b, 0] + 1j * Psi[d, b, 1]
        # scanning phase with subpixel precision
        v = cm.exp(-2j * cm.pi * (qB[0, b] * rho0 + qB[1, b] * rho1)) * Psic
        out[b, d, k, 0] = v.real
        out[b, d, k, 1] = v.imag

# Cell
@cuda.jit
def phase_factor_kernelKB(Psi, rho, qB, out):
    """
    Calculate the phase factors (due to beam scan) probe wave function so that
    the probe is scanned to the correct place for each diffraction pattern

    :param Psi:         B x 2
        Probe wave functions Fourier coefficient for each beam to be mutliplied
        by phase factor to account for beam scan position
    :param rho:         K x 2
        Probe positions in pixels
    :param qB:          2 x B
        Fourier space coordinates of the beams
    :param out:         K x B x 2
        Phase factors output
    :return: scanning phases for all defoc, beams, positions
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K, B, c = out.shape
    N = B * K

    b = n // (K)
    k = (n - b * K)

    if n < N:
        rho0 = rho[k, 0]
        rho1 = rho[k, 1]
        Psic = Psi[b, 0] + 1j * Psi[b, 1]
        # scanning phase with subpixel precision
        v = cm.exp(-2j * cm.pi * (qB[0, b] * rho0 + qB[1, b] * rho1)) * Psic
        out[k, b, 0] = v.real
        out[k, b, 1] = v.imag

# Cell
@cuda.jit
def smatrix_forward_kernel_fast_full4(S, phase_factors, r, r_min, out):
    """

    :param S:               B x NY x NX

    :param phase_factors:   B x D x K x 2
    :param r:               D x K x 2
    :param out:             D x K x MY x MX
    :param out_strides: (4,)
    :return: exit waves in out
    """
    k, my, mx = cuda.grid(3)
    D, K, MY, MX, _ = out.shape
    B = S.shape[0]

    if k < K and my < MY and mx < MX:
        for d in range(D):
            accum_real = 0.
            accum_imag = 0.
            for b in range(B):
                rho0 = int(r[d, k, 0] - r_min[0])
                rho1 = int(r[d, k, 1] - r_min[1])
                S_b_real = S[b, rho0 + my, rho1 + mx, 0]
                S_b_imag = S[b, rho0 + my, rho1 + mx, 1]
                a = S_b_real
                c = S_b_imag
                u = phase_factors[b, d, k, 0]
                v = phase_factors[b, d, k, 1]

                accum_real += a * u - c * v
                accum_imag += c * u + a * v

            out[d, k, my, mx, 0] = accum_real
            out[d, k, my, mx, 1] = accum_imag

@cuda.jit
def split_kernel(S, r, out):
    """

    :param S: B x NY x NX x 2
    :param r: K x2
    :param out: K x MY x MX x 2
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K, MY, MX, B, _ = out.shape
    N = K * MY * MX * B

    k = n // (MY * MX * B)
    my = (n - k * MY * MX * B) // (MX * B)
    mx = (n - k * MY * MX * B - my * MX * B) // B
    b = (n - k * MY * MX * B - my * MX * B - mx * B)

    if n < N:
        y = r[k, 0]
        x = r[k, 1]
        out[k, my, mx, b, 0] = S[b, y + my, x + mx, 0]
        out[k, my, mx, b, 1] = S[b, y + my, x + mx, 1]

@cuda.jit
def split_kernel4(S, r, out):
    """

    :param S:   B x NY x NX
    :param r:   K x 2
    :param out: B x K x MY x MX
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    B, K, MY, MX = out.shape
    N = K * MY * MX * B

    b = n // (MY * MX * K)
    k = (n - b * MY * MX * K) // (MX * MY)
    my = (n - b * MY * MX * K - k * MX * MY) // MX
    mx = (n - b * MY * MX * K - k * MX * MY - my * MX)

    if n < N:
        y = r[k, 0]
        x = r[k, 1]
        out[b, k, my, mx] = S[b, y + my, x + mx]

@cuda.jit
def split_kernel5(S, r, out):
    """

    :param S: B x NY x NX x 2
    :param r: K x 2
    :param out: K x B x MY x MX x 2
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K, B, MY, MX, _ = out.shape
    N = K * MY * MX * B

    k = n // (MY * MX * B)
    b = (n - k * MY * MX * B) // (MX * MY)
    my = (n - k * MY * MX * B - k * MX * MY) // MX
    mx = (n - k * MY * MX * B - k * MX * MY - my * MX)

    if n < N:
        y = r[k, 0]
        x = r[k, 1]
        out[k, b, my, mx, 0] = S[b, y + my, x + mx, 0]
        out[k, b, my, mx, 1] = S[b, y + my, x + mx, 1]

@cuda.jit
def split_kernel2(S, r, out):
    """

    :param S: B x NY x NX x 2
    :param r: K x2
    :param out: K x MY x MX x 2
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K, MY, MX, B, _ = out.shape
    N = K * MY * MX * B

    k = n // (B)
    b = (n - k * B)

    if n < N:
        for my in range(MY):
            for mx in range(MX):
                y = r[k, 0]
                x = r[k, 1]
                out[k, my, mx, b, 0] = S[b, y + my, x + mx, 0]
                out[k, my, mx, b, 1] = S[b, y + my, x + mx, 1]

@cuda.jit
def split_kernel3(S, r, out):
    """
    :param S: B x NY x NX x 2
    :param r: K x2
    :param out: K x MY x MX x 2
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K, MY, MX, B, _ = out.shape
    N = K * MY * MX * B

    k = n // (MY * B)
    my = (n - k * MY * B) // (B)
    b = (n - k * MY * B - my * B)

    if n < N:
            for mx in range(MX):
                y = r[k, 0]
                x = r[k, 1]
                out[k, my, mx, b, 0] = S[b, y + my, x + mx, 0]
                out[k, my, mx, b, 1] = S[b, y + my, x + mx, 1]

# Cell
@cuda.jit
def overlap_kernel_real2(r, z, out):
    """

    :param r: K x 2
    :param z: BB x K x MY x MX
    :param out: BB x NY x NX
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K = r.shape[0]
    BB, MY, MX = z.shape
    N = BB * K * MY * MX

    bb = n // (K * MY * MX)
    k = (n - bb * (K * MY * MX)) // (MY * MX)
    my = (n - bb * (K * MY * MX) - k * MY * MX) // MX
    mx = (n - bb * (K * MY * MX) - k * MY * MX - my * MX)

    if n < N:
        y = r[k, 0]
        x = r[k, 1]
        val = z[bb, my, mx]
        cuda.atomic.add(out, (bb, y + my, x + mx), val)

# Cell
@cuda.jit
def psi_denom_kernel(r, t, out):
    """

    :param r: K x 2
    :param t: BB x NY x NX
    :param out: BB x MY x MX
    :return:
    """

    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K = r.shape[0]
    BB, MY, MX = out.shape
    N = BB * K * MY * MX

    bb = n // (K * MY * MX)
    k = (n - bb * (K * MY * MX)) // (MY * MX)
    my = (n - bb * (K * MY * MX) - k * (MY * MX)) // MX
    mx = (n - bb * (K * MY * MX) - k * (MY * MX) - my * MX)

    if n < N:
        y = r[k, 0]
        x = r[k, 1]
        val = abs(t[bb, y + my, x + mx]) ** 2
        cuda.atomic.add(out, (bb, my, mx), val)

# Cell
@cuda.jit
def psi_kernel(r, t, z, out):
    """

    :param r:   K x 2
    :param t:   BB x NY x NX
    :param z:   K x MY x MX
    :param out: BB x MY x MX
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    K = r.shape[0]
    MY, MX = out.shape
    N = K * MY * MX

    k = (n // (MY * MX))
    my = ((n - k * (MY * MX)) // MX)
    mx = ((n - k * (MY * MX) - my * MX))

    if n < N:
        y = r[k, 0]
        x = r[k, 1]
        t_conj = t[y + my, x + mx].conjugate()
        val = t_conj * z[k, my, mx]
        cuda.atomic.add(out.real, (my, mx), val.real)
        cuda.atomic.add(out.imag, (my, mx), val.imag)

# Cell
@cuda.jit
def A_realspace_kernel(r, t, psi, out):
    """

    :param r:   K x 2
    :param t:   B x NY x NX
    :param psi: B x K x MY x MX
    :param out: K x MY x MX
    :return:
    """
    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    B, K, MY, MX, _ = psi.shape
    N = K * MY * MX

    k = n // (MY * MX)
    my = (n - k * (MY * MX)) // MX
    mx = (n - k * (MY * MX) - my * MX)

    if n < N:
        for bb in range(B):
            y = r[k, 0]
            x = r[k, 1]

            # val = t[bb, y + my, x + mx] * psi[bb, k, my, mx]
            # cuda.atomic.add(out.real, (k, y + my, x + mx), val.real)
            # cuda.atomic.add(out.imag, (k, y + my, x + mx), val.imag)
            #
            a = t[bb, y + my, x + mx, 0]
            b = t[bb, y + my, x + mx, 1]
            u = psi[bb, k, my, mx, 0]
            v = psi[bb, k, my, mx, 1]

            val_real = a * u - b * v
            val_imag = b * u + a * v

            cuda.atomic.add(out, (k, my, mx, 0), val_real)
            cuda.atomic.add(out, (k, my, mx, 1), val_imag)

# Cell
@cuda.jit
def AtF2_kernel(z, psi, r, out):
    """

    :param z:   K x MY x MX
    :param psi: B x K x MY x MX
    :param r:   K x 2
    :param out: B x NY x NX
    :return:
    """

    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    B, K, MY, MX = psi.shape
    N = B * K * MY * MX

    bb = (n // (MY * MX * K))
    k = (n - bb * (MY * MX * K)) // (MY * MX)
    my = (n - bb * (MY * MX * K) - k * (MY * MX)) // MX
    mx = (n - bb * (MY * MX * K) - k * (MY * MX) - my * MX)

    if n < N:
        y = r[k, 0]
        x = r[k, 1]
        val = psi[bb, k, my, mx].conjugate() * z[k, my, mx]
        cuda.atomic.add(out.real, (bb, y + my, x + mx), val.real)
        cuda.atomic.add(out.imag, (bb, y + my, x + mx), val.imag)

# Cell
@cuda.jit
def dS_kernel(z, z_old, psi, psi_int, psi_int_max, alpha, r, out):
    """
    :param z:           K x MY x MX
    :param z_old:       K x MY x MX
    :param psi:         B x K x MY x MX
    :param psi_int:     B x K x MY x MX
    :param psi_int_max: B x K
    :param r:           K x 2
    :param out:         B x NY x NX
    :param alpha:       float
    """

    n = cuda.blockIdx.x * cuda.blockDim.x + cuda.threadIdx.x
    B, K, MY, MX = psi.shape
    N = B * K * MY * MX

    bb = (n // (MY * MX * K))
    k = (n - bb * (MY * MX * K)) // (MY * MX)
    my = (n - bb * (MY * MX * K) - k * (MY * MX)) // MX
    mx = (n - bb * (MY * MX * K) - k * (MY * MX) - my * MX)

    if n < N:
        y = r[k, 0]
        x = r[k, 1]

        denom = (1-alpha)*psi_int[bb, k, my, mx] + alpha * psi_int_max[bb, k]
        val = (psi[bb, k, my, mx].conjugate() * (z[k, my, mx] )) / denom
        # val = (psi[bb, k, my, mx].conjugate() * (z[k, my, mx]))
        # - z_old[k, my, mx]
        cuda.atomic.add(out, (bb, y + my, x + mx,0), val.real)
        cuda.atomic.add(out, (bb, y + my, x + mx,1), val.imag)